## This code is used to calculate the gene_based interaction tests
## use the methods described in the following paper:
## Gene-based testing of interactions in association studies of quantitative traits. PLoS Genet.
## Ma L, Clark AG, Keinan A.  9:e1003321 (2013)
## Changed by Yingjie GUO

## log means log sum rather than product 

##################################################

debug = TRUE
display = FALSE
reps=0 # Maybe offer reps as a parameter? Defaul = 1000

##################################################


if(debug){
  
  ## Manually assign the args for debugging
  print("--------------------------------")
  print("Currently in the debugging mode!")
  print("--------------------------------")
  gene1 = "SET1"
  gene2 = "SET2"
  pvaluesFile = paste0("snp_pair_results1.txt")
  corrFile = paste0("snp_pair_results1.txt.summary.cori")
  outputFile = "result_bt_log_debug.txt"
  ## The filePath in Linux
  filePath = "./" # current path
  ## The filePath in Mac
  #filePath = "/Users/SenZHANG/Dropbox/PhD_Year_3/Research_Topics/Community_Detection/Experiments/XWAS/example_Sen"
  # corrMethod = "chol"
  
  setwd(filePath)
  
  #if(file.exists(outputFile)){
  #  file.remove(outputFile)
  #}
  
} else{
  
  ## Need to add Rscript.exe into system's PATH to run this script in cmd line
  args <- commandArgs(TRUE)
  pvaluesFile = args[1]
  corrFile = args[2]
  gene1=args[3]
  gene2=args[4]
  outputFile = args[5]
  # From python: os.getcwd()
  filePath = args[6] #"D:\\Dropbox\\PhD_Year_3\\Research_Topics\\Community_Detection\\Experiments\\XWAS\\example_Sen"
  # corrMethod = args[7] # Now determined by the rank of the covariance matrix! Full rank: 'chol' Low rank: 'svd'
  setwd(filePath)
  
}

## Loading libraries in Windows
# newPackLoc <- "C:\\Users\\szhangat\\Documents\\R\\win-library\\3.1"
# library(mvtnorm, lib.loc = newPackLoc, verbose =  FALSE)
# library(corpcor, lib.loc = newPackLoc, verbose = FALSE)

## Loading libraries in Mac
library(mvtnorm)
library(corpcor)


if (file.info(corrFile)$size == 0){
  print("Quitting gene_based_inter_bt.R, your correlation file is empty!")
  quit(save = "no", status = 0, runLast = FALSE)
}

#################################################################################################
#################################################################################################
#################################################################################################
# function to get the truncated tail p-value
mytail=function(x) {
  temp=sort(x[x<myth])
  m=length(temp)
  if (m>0){
    ts=0
    for (i in 1:m) ts=ts+1-temp[i]*(length(x)+1)/i
    return(ts/length(x)) }
  else {return(-9999)}
}

# function to obtain the truncated prod p-value
myprod=function(x) {
  temp=sort(x[x<myth])
  m=length(temp)
  if (m>0){
    return(prod(temp))
  }
  else {
    return(1)
  }
}

# function to obtain the sum of log truncated p-value
mylogsum = function(x){
  temp = sort(x[x<myth])
  m = length(temp)
  if(m>0){
    return(sum(log(temp)))
  }
  else{
    return(1)
  }
}

# function to get the gene-based p-value
gbpvalue = function(reps,numsnps,co,pv) {
  # convert random statistics to p-values.
  obstailpv = mytail(pv)
  simpvtail = 0
  
  obsprodpv = myprod(pv)
  simpvprod = 0
  
  obslogsumpv = mylogsum(pv)
  simpvlogsum = 0
  
  numiter = reps/1000
  for (i in 1:numiter){
    currsimpv = pnorm(rmvnorm(1000,mean=rep(0,numsnps),sigma=co,method=corrMethod),lower.tail=FALSE)*2
    simpvtail=simpvtail+sum(apply(currsimpv,1,mytail)>=obstailpv)
    simpvprod=simpvprod+sum(apply(currsimpv,1,myprod)<=obsprodpv)
    simpvlogsum=simpvlogsum+sum(apply(currsimpv,1,mylogsum)<=obslogsumpv)
  }
  list(simpvtail/reps,simpvprod/reps, simpvlogsum/reps)
}

#################################################################################################
###############################END OF FUNCTIONS##################################################
#################################################################################################

#-----------------------------------------------------------------------------------------------#
## Run the gene_based test


myth=0.05

pv <- as.matrix(read.table(pvaluesFile,header=T))[,6]
pv <- as.numeric(pv)

cori=as.matrix(read.table(corrFile,header=FALSE,sep="\t"))

if(qr(cori)$rank < dim(cori)[1]){
  corrMethod = 'svd'
} else{
  corrMethod = 'chol'
}

num.pv <- length(pv)
ptemp <- rep(NA,num.pv)


if(debug && display){
  # For debugging, can be commented
  print(paste0('The dimension of the correlation matrix: ', dim(cori)[1]))
  print(paste0('The rank of the correlation matrix: ', qr(cori)$rank))
  print(paste0("The method used for corr matrix: ", corrMethod))
  library(fields)
  image.plot(cori) # Don't enable this in Mac
}

# calc pmin
t1 <- Sys.time()
minth=qnorm(min(pv)/2)
pmin=1-pmvnorm(lower=minth,upper=-minth,mean=rep(0, num.pv),corr=cori)
t2 <- Sys.time()
print(t2-t1)

# calc pgate
# me = pmin/min(pv) # Previously use me as sum(keff)
# ptemp[1] = pmin
keff = rep(NA, num.pv)
keff[1] = 1
myo=order(pv) # myo should be in increasing order! (Also the default of order)
for (i in 2:num.pv) {
  # In python, import numpy as np, then -1.31*np.log10(0.05)=1.704349
  # keff=keff+sqrt(abs(1-max(abs(cori[myo[1:(i-1)],myo[i]]))^1.71))
  keff[i]=sqrt(abs(1-max(abs(cori[myo[1:(i-1)],myo[i]]))^(-1.31*log10(myth))))
}
me = sum(keff)
for (i in 1:num.pv) {
  ptemp[i] = me*pv[myo[i]]/keff[i]
}
pgate=min(ptemp)

# Following VEGAS with adaptive permutations
xx=gbpvalue(reps,num.pv,cori,pv)
if (xx[1] < 0.1 || xx[2] < 0.1 || xx[3] < 0.1) {
  reps=10000
  print(paste0("current reps = ", reps, '...'))
  xx=gbpvalue(reps,num.pv,cori,pv)
  
  ## Comment this part for simulations for type-1-err to speed up
  # if(xx[1]<0.001||xx[2]<0.001||xx[3]<0.001){
  #   reps=1e6
  #   print(paste0("current reps = ", reps, '...'))
  #   xx=gbpvalue(reps,num.pv,cori,pv)
  # }

}

# write the result
# format: gene1 gene2 pmin  pgate ptts  pprod plogsum 
# phr = sprintf("%s\t%s\t%11s\t%11s\t%11s\t%11s\t%11s", "gene1","gene2","pmin","pgate","ptts","pprod","plogsum")
# write(phr,file=outputFile,append=TRUE)

phr = sprintf("%s\t%s\t%11.5e\t%11.5e\t%11.5e\t%11.5e\t%11.5e", gene1,gene2,pmin,pgate,xx[1],xx[2],xx[3])
write(phr,file=outputFile,append=TRUE)


